'''
# Some Basic Constants in Physics
## Avaiable Constants:\n
    Gravitational Constant: G=6.67259e-11\n
    Speed of light: c=299792458\n    
    Light year: Ly=9.46e12\n
    Astronomical Unit: AU=1.496e12\n
    Mass of planets:\n
        M_sun=1.989e30\n
        M_earth=5.974e24\n    
    Plank Constant: Plank_h=6.6262e-34\n
    Charge of electron: Qe=1.6022e-19\n
    Mass of basic particles:\n
        Me=9.1095e-31\n
        Mp=1.6726e-27\n
        Mn=1.67493e-27\n
    Avogadro Constant: Na=6.022e26\n
    Steven-Boltzman Constant: Sigma=5.67e-8\n    
    Electric Vault: eV=1.6022e-19\n    
    Radius of planets:\n
        R_earth=6371393\n
        R_sun=696300000\n    
    Spin of planets:\n
        S_earth=[0,0,7.292e-5]\n
    permeability:\n
        mu0=4*math.pi*1e-7

@AUTHOR: Yufeng Wang
@DATE: 2022-6
@LICENSE: GPL-V3 license

CopyRight (C) Yufeng-Wang/Airscker, 2022-6

More information about this module please see: https://airscker.github.io/Purple-Micro-Star/
'''

import math
# Gravity
G=6.67259e-11
# Speed of light
c=299792458
# Light year
Ly=9.46e12
# Celestial distance
AU=1.496e12
# Mass of planets
M_sun=1.989e30
M_earth=5.974e24
# Plank const
Plank_h=6.6262e-34
# Q of electron
Qe=1.6022e-19
# Mass of basic particles
Me=9.1095e-31
Mp=1.6726e-27
Mn=1.67493e-27
# Avogadro constant
Na=6.022e26
# Steven-Boltzman const
Sigma=5.67e-8
# Electric Vault
eV=1.6022e-19
# Radius of planet
R_earth=6371393
R_sun=696300000
# Spin of planet
S_earth=[0,0,7.292e-5]
# Megnatic
mu0=4*math.pi*1e-7
